/* Implements a camera reference system with some transformations.
The reference system have one origin (relative to universe reference
system) and 3 vectors (I, J, K) with module 1.
Example:
             J
            |
            |  / K
            | /
            |/
            *------
             O     I

WARNING:
- after change floating point base of the reference system you must update
  the fix point base by calling CCameraSystem::mComputeFixPointBase.
*/



#include <string.h>

#include <math.h>

#include "defs.h"

#include "node.h"

#include "nodever.h"

#include "nodexy.h"

#include "list.h"

#include "camsys.h"



/* constructor - init default reference system
**********************************************************/
CCameraSystem::CCameraSystem ()
{
   mCreate ();
}



/* constructor - init reference system with other one.
**********************************************************/
CCameraSystem::CCameraSystem ( CCameraSystem& oCS )
{
   mCreate ();
   mSetObj ( oCS );
}



/* destructor
**********************************************************/
CCameraSystem::~CCameraSystem ()
{
}



/* Init reference system and lockup tables
**********************************************************/
void CCameraSystem::mCreate ()
{
   m_nDistCP = DIST_CP;
   m_fRadius = RAD;   
   m_fSpeed = SPEED;

   mResetPos ();
   mResetBase ();

   mComputeFixPointBase ();

   // REVIEW (criar CTrans e por lah)
   // sin and cos lockup tables
   for ( INT ncD = 0; ncD < ROUND; ncD ++ )  {
      // fix point 16.16
      m_anSin [ ncD ] = (INT) ( sin ( ncD * PI / ROUND * 2 ) * 65536 );
      m_anCos [ ncD ] = (INT) ( cos ( ncD * PI / ROUND * 2 ) * 65536 );
      // floating point
      m_afSin [ ncD ] = sin ( ncD * PI / ROUND * 2 );
      m_afCos [ ncD ] = cos ( ncD * PI / ROUND * 2 );
   }
}



/* Destroy object
**********************************************************/
void CCameraSystem::mDestroy ()
{
}



/* Set position to default value: space origin
**********************************************************/
void CCameraSystem::mResetPos ()
{
   m_stPos.fX = m_stPos.fY = 0;
   m_stPos.fZ = 0;//- m_nDistCP;   // change projection plan to camera location
}



/* Set reference system base to default value.
**********************************************************/
void CCameraSystem::mResetBase ()
{
   // set I
   m_stI.fX = 1;
   m_stI.fY = 0;
   m_stI.fZ = 0;
   // set J
   m_stJ.fX = 0;
   m_stJ.fY = 1;
   m_stJ.fZ = 0;
   // set K
   m_stK.fX = 0;
   m_stK.fY = 0;
   m_stK.fZ = -1;

   mComputeFixPointBase ();
}



/* Init reference system with a copy of other object
(like CCameraSystem::constructor and CCameraSystem::operator =)
**********************************************************/
void CCameraSystem::mSetObj ( CCameraSystem& oCS )
{
   // set position
   m_stPos.fX = oCS.m_stPos.fX;
   m_stPos.fY = oCS.m_stPos.fY;
   m_stPos.fZ = oCS.m_stPos.fZ;
   // set I
   m_stI.fX = oCS.m_stI.fX;
   m_stI.fY = oCS.m_stI.fY;
   m_stI.fZ = oCS.m_stI.fZ;
   // set J
   m_stJ.fX = oCS.m_stJ.fX;
   m_stJ.fY = oCS.m_stJ.fY;
   m_stJ.fZ = oCS.m_stJ.fZ;
   // set K
   m_stK.fX = oCS.m_stK.fX;
   m_stK.fY = oCS.m_stK.fY;
   m_stK.fZ = oCS.m_stK.fZ;
   // distance of projection center and projection plane
   m_nDistCP = oCS.m_nDistCP;
   // radius
   m_fRadius = oCS.m_fRadius;
   // move speed
   m_fSpeed = oCS.m_fSpeed;

   mComputeFixPointBase ();
}



/* like CCameraSystem::constructor and CCameraSystem::mSetObj
**********************************************************/
void CCameraSystem::operator = ( CCameraSystem& oCS )
{
   mSetObj ( oCS );
}



/* Transform floating point base to fix point 16.16.
Call after change floating point base.
**********************************************************/
void CCameraSystem::mComputeFixPointBase ()
{
   m_stFixI.nX = (INT) ( m_stI.fX * 65536 );
   m_stFixI.nY = (INT) ( m_stI.fY * 65536 );
   m_stFixI.nZ = (INT) ( m_stI.fZ * 65536 );
   m_stFixJ.nX = (INT) ( m_stJ.fX * 65536 );
   m_stFixJ.nY = (INT) ( m_stJ.fY * 65536 );
   m_stFixJ.nZ = (INT) ( m_stJ.fZ * 65536 );
   m_stFixK.nX = (INT) ( m_stK.fX * 65536 );
   m_stFixK.nY = (INT) ( m_stK.fY * 65536 );
   m_stFixK.nZ = (INT) ( m_stK.fZ * 65536 );
}



/* Transform this camera system acording the given camera system.
Update camera position and orientation parameters.
**********************************************************/
void CCameraSystem::mTransform ( CCameraSystem& oC )
{
   // transform position
   m_stPos.fX += oC.m_stPos.fX;
   m_stPos.fY += oC.m_stPos.fY;
   m_stPos.fZ += oC.m_stPos.fZ;
   /*// transform I
   float
      fXa = m_stI.fX * oC.m_stI.fX + m_stI.fY * oC.m_stI.fY + m_stI.fZ * oC.m_stI.fZ,
      fYa = m_stI.fX * oC.m_stJ.fX + m_stI.fY * oC.m_stJ.fY + m_stI.fZ * oC.m_stJ.fZ,
      fZa = m_stI.fX * oC.m_stK.fX + m_stI.fY * oC.m_stK.fY + m_stI.fZ * oC.m_stK.fZ;
   m_stI.fX = fXa;
   m_stI.fY = fYa;
   m_stI.fZ = fZa;
   // transform J
   fXa = m_stJ.fX * oC.m_stI.fX + m_stJ.fY * oC.m_stI.fY + m_stJ.fZ * oC.m_stI.fZ;
   fYa = m_stJ.fX * oC.m_stJ.fX + m_stJ.fY * oC.m_stJ.fY + m_stJ.fZ * oC.m_stJ.fZ;
   fZa = m_stJ.fX * oC.m_stK.fX + m_stJ.fY * oC.m_stK.fY + m_stJ.fZ * oC.m_stK.fZ;
   m_stJ.fX = fXa;
   m_stJ.fY = fYa;
   m_stJ.fZ = fZa;
   // transform K
   fXa = m_stK.fX * oC.m_stI.fX + m_stK.fY * oC.m_stI.fY + m_stK.fZ * oC.m_stI.fZ;
   fYa = m_stK.fX * oC.m_stJ.fX + m_stK.fY * oC.m_stJ.fY + m_stK.fZ * oC.m_stJ.fZ;
   fZa = m_stK.fX * oC.m_stK.fX + m_stK.fY * oC.m_stK.fY + m_stK.fZ * oC.m_stK.fZ;
   m_stK.fX = fXa;
   m_stK.fY = fYa;
   m_stK.fZ = fZa;*/

   // transform I
   float
      fXa = m_stI.fX * oC.m_stI.fX + m_stI.fY * oC.m_stI.fX + m_stI.fZ * oC.m_stI.fX,
      fYa = m_stI.fX * oC.m_stI.fY + m_stI.fY * oC.m_stI.fY + m_stI.fZ * oC.m_stI.fY,
      fZa = m_stI.fX * oC.m_stI.fZ + m_stI.fY * oC.m_stI.fZ + m_stI.fZ * oC.m_stI.fZ;
   m_stI.fX = fXa;
   m_stI.fY = fYa;
   m_stI.fZ = fZa;
   // transform J
   fXa = m_stJ.fX * oC.m_stJ.fX + m_stJ.fY * oC.m_stJ.fX + m_stJ.fZ * oC.m_stJ.fX;
   fYa = m_stJ.fX * oC.m_stJ.fY + m_stJ.fY * oC.m_stJ.fY + m_stJ.fZ * oC.m_stJ.fY;
   fZa = m_stJ.fX * oC.m_stJ.fZ + m_stJ.fY * oC.m_stJ.fZ + m_stJ.fZ * oC.m_stJ.fZ;
   m_stJ.fX = fXa;
   m_stJ.fY = fYa;
   m_stJ.fZ = fZa;
   // transform K
   fXa = m_stK.fX * oC.m_stK.fX + m_stK.fY * oC.m_stK.fX + m_stK.fZ * oC.m_stK.fX;
   fYa = m_stK.fX * oC.m_stK.fY + m_stK.fY * oC.m_stK.fY + m_stK.fZ * oC.m_stK.fY;
   fZa = m_stK.fX * oC.m_stK.fZ + m_stK.fY * oC.m_stK.fZ + m_stK.fZ * oC.m_stK.fZ;
   m_stK.fX = fXa;
   m_stK.fY = fYa;
   m_stK.fZ = fZa;

   mComputeFixPointBase ();
}



/* Apply the reference system base to a point stPoint
**********************************************************/
void CCameraSystem::mApplyBase ( tagfXYZ& stPoint )
{
   float
      fXa = stPoint.fX * m_stI.fX + stPoint.fY * m_stI.fY + stPoint.fZ * m_stI.fZ,
      fYa = stPoint.fX * m_stJ.fX + stPoint.fY * m_stJ.fY + stPoint.fZ * m_stJ.fZ,
      fZa = stPoint.fX * m_stK.fX + stPoint.fY * m_stK.fY + stPoint.fZ * m_stK.fZ;
   stPoint.fX = fXa;
   stPoint.fY = fYa;
   stPoint.fZ = fZa;
}



// REVIEW
/* Rotate reference system base using the transpose matrix
**********************************************************/
void CCameraSystem::mRotateTransBase ( BYTE yAX, BYTE yAY, BYTE yAZ )
{
   mRotate ( yAX, yAY, yAZ, m_stI.fX, m_stJ.fX, m_stK.fX );
   mRotate ( yAX, yAY, yAZ, m_stI.fY, m_stJ.fY, m_stK.fY );
   mRotate ( yAX, yAY, yAZ, m_stI.fZ, m_stJ.fZ, m_stK.fZ );

   mComputeFixPointBase ();
}



/* Rotate reference system base
**********************************************************/
void CCameraSystem::mRotateBase ( BYTE yAX, BYTE yAY, BYTE yAZ )
{
   mRotate ( yAX, yAY, yAZ, m_stI.fX, m_stI.fY, m_stI.fZ );
   mRotate ( yAX, yAY, yAZ, m_stJ.fX, m_stJ.fY, m_stJ.fZ );
   mRotate ( yAX, yAY, yAZ, m_stK.fX, m_stK.fY, m_stK.fZ );

   mComputeFixPointBase ();
}



/* Rotate a point
**********************************************************/
void CCameraSystem::mRotate ( BYTE yAX, BYTE yAY, BYTE yAZ,
                              float& fX, float& fY, float& fZ )
{
   // rotated temp components
   float
      fXAux,
      fYAux,
      fZAux;

   fXAux = fX    * * ( m_afCos + yAZ ) + fY    * * ( m_afSin + yAZ );
   fYAux = fY    * * ( m_afCos + yAZ ) - fX    * * ( m_afSin + yAZ );
   fY    = fYAux * * ( m_afCos + yAX ) + fZ    * * ( m_afSin + yAX );
   fZAux = fZ    * * ( m_afCos + yAX ) - fYAux * * ( m_afSin + yAX );
   fX    = fXAux * * ( m_afCos + yAY ) + fZAux * * ( m_afSin + yAY );
   fZ    = fZAux * * ( m_afCos + yAY ) - fXAux * * ( m_afSin + yAY );
}
